;; Generated package description from corfu.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "corfu" "2.6" "COmpletion in Region FUnction" '((emacs "29.1") (compat "30")) :commit "81af89b43aceb5714fcd21daabcfd4b77008e22e" :authors '(("Daniel Mendler" . "mail@daniel-mendler.de")) :maintainer '("Daniel Mendler" . "mail@daniel-mendler.de") :keywords '("abbrev" "convenience" "matching" "completion" "text") :url "https://github.com/minad/corfu")
